/*
 *    NoiascaLedControl.h - A library for controling LEDss with a MAX7219/MAX7221
 *    
 *    noiasca
 *    https://werner.rothschopf.net/201904_arduino_ledcontrol_max7219.htm
 * 
 *    Permission is hereby granted, free of charge, to any person
 *    obtaining a copy of this software and associated documentation
 *    files (the "Software"), to deal in the Software without
 *    restriction, including without limitation the rights to use,
 *    copy, modify, merge, publish, distribute, sublicense, and/or sell
 *    copies of the Software, and to permit persons to whom the
 *    Software is furnished to do so, subject to the following
 *    conditions:
 * 
 *    This permission notice shall be included in all copies or 
 *    substantial portions of the Software.
 * 
 *    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 *    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 *    OTHER DEALINGS IN THE SOFTWARE.
 
 *    Version
 *    2021-02-28 noiasca 1.2.4  cleanup name conflict with NanoBLE33
 *    2020-06-22 noiasca 1.2.1  fix VT table, removed B10000000 constants
 *    2020-12-08 noiasca        split into abstract base class and hw implementation
 *    2020-08-08 noiasca        no hardware calls in constructor, begin() method has to be called separatly, soft and hardware SPI splitted
 *    2020-03-07 noiasca        add hardware SPI
 *    2020-03-06 noiasca        rebuild int to uint8_t
 *    2019-04-25 noiasca        inherit print class
 *    2007       Eberhard Fahle 
 *
 */

 /*  Extensions by noiasca:
 
 The modules/devices digits are addressed as following:
 
 device   0                 1                 2                       7
 addr     0 0 0 0 0 0 0 0   1 1 1 1 1 1 1 1   2 2 2 2 2 2 2 2         7 7 7 7 7 7 7 7
 digit    7 6 5 4 3 2 1 0   7 6 5 4 3 2 1 0   7 6 5 4 3 2 1 0   ...   7 6 5 4 3 2 1 0
 
 position 0 1 2 3 4 5 6 7   8 9 10            16                      56            63

 offene Punkte 
 - EndOfDevice                          0                                  1                   2             3
 - Bei "Adress-Ende":                   Beginne Am Anfang dieser Adresse | Beginne bei 0 / 7 | Shift Links | Weiter in nächste Adresse |  
 - Bei Ende letzte Adrese (Zeilenende): Beginne Am Anfang dieser Adresse | Beginne bei 0 / 7 | Shift Links |
 - Schreiben über das Ende eines Devices (über das Ende der Device-Kette)ermöglichen (Position 254??)

 The files NoiascaLedControlCommon contains an abstract class.
 This NoiascaLedControl_base class contains all shared variables and methods.
 The NoiascaLedControl inherits from NoiascaLedControl_base and is for Software bit bang
 The NoiascaLedControlSpi inherits from NoiascaLedControl_base and is for HW SPI
 
 NoiascaLedControl_base
  <--NoiascaLedControl
  <--NoiascaLedControlSPI
 */
 
#pragma once
#define NoiascaLedControl_h

#if defined(__AVR__)
#include <avr/pgmspace.h>
#elif defined(ESP8266)
#include <pgmspace.h>
#endif

#if (ARDUINO >= 100)
#include <Arduino.h>
#else
//#include <WProgram.h>
#endif

#include "Print.h"
#include "NoiascaLedControlCommon.h"

/** \brief class for MAX7219/MAX7221 (software bitbang)

    this class should be used if you can't use Hardware SPI.
*/
class LedControl : public LedControl_base {
  protected:
    const uint8_t dataPin;        /* Data is shifted out of this pin, is now in the respective HW classes. former SPI_MOSI*/
    const uint8_t clkPin;         /* The clock is signaled on this pin. former SPI_CLK */
    const uint8_t csPin;          /* This one is driven LOW for chip selection. former SPI_CS */
    void spiTransfer(uint8_t addr, uint8_t opcode, uint8_t data);     /* Send out a single command to the device */

  public:
     /** \brief Create a new MAX7219/MAX7221 object for Hardware SPI
      
      \param  dataPin		  pin on the Arduino where data gets shifted out
      \param  clkPin		  pin for the clock
      \param  csPin pin   for selecting the device 
      \param  numDevices  maximum number of devices that can be controlled
     */
    LedControl(uint8_t dataPin, uint8_t clkPin, uint8_t csPin, uint8_t numDevices = 1);
    
     /** \brief initialise hardware
     
       call this method in your setup().
       initialise hardware - sets the pinMode for all needed GPIO        
     */       
    void begin();
};
