/*
 *    NoiascaLedControl.h - A library for controling Leds with a MAX7219/MAX7221
 *    
 *    noiasca
 *    https://werner.rothschopf.net/201904_arduino_ledcontrol_max7219.htm
 * 
 *    Permission is hereby granted, free of charge, to any person
 *    obtaining a copy of this software and associated documentation
 *    files (the "Software"), to deal in the Software without
 *    restriction, including without limitation the rights to use,
 *    copy, modify, merge, publish, distribute, sublicense, and/or sell
 *    copies of the Software, and to permit persons to whom the
 *    Software is furnished to do so, subject to the following
 *    conditions:
 * 
 *    This permission notice shall be included in all copies or 
 *    substantial portions of the Software.
 * 
 *    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 *    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 *    OTHER DEALINGS IN THE SOFTWARE.
 
 *    Version
 *    This file only:
 *    2022-03-23 noiasca        removed globals and #defines
 *    2022-02-28 noiasca        added comments for doxygen
 *    2020-12-08 noiasca        clean up classes, no duplicate code
 */
 
 /*! 
 
  \mainpage Some words to the Noiasca LED Control Library
 
  \section intro_sec Introduction
  The "Noiasca LED Control" is an Arduino Library to control
  MAX7219/MAX7221 seven segment LED displays.
  
  The library is an improved fork of "LED Control" by Eberhard Fahle and
  inherits from Print.h. 
  Furthermore it supports communication on any pin (software bitbang) 
  and faster communication on HW SPI.
    
  \section example_sec Examples
  There are several examples please use the hello world for the beginning.
  
  \section install_sec Install the library
  Download the library from https://werner.rothschopf.net/201904_arduino_ledcontrol_max7219.htm
  
  In the Arduino IDE use the Menu <br>
   Sketch / Include Library / Add .ZIP Library <br>
  to install the library.
 */
 
/*!
@file 
Contains base class and global enumerations
*/

 /*  Extensions by noiasca:
 
 The modules/devices digits are addressed as following:
 
 device   0                 1                 2                       7
 addr     0 0 0 0 0 0 0 0   1 1 1 1 1 1 1 1   2 2 2 2 2 2 2 2         7 7 7 7 7 7 7 7
 digit    7 6 5 4 3 2 1 0   7 6 5 4 3 2 1 0   7 6 5 4 3 2 1 0   ...   7 6 5 4 3 2 1 0
 
 position 0 1 2 3 4 5 6 7   8 9 10            16                      56            63

 offene Punkte 
 - EndOfDevice                          0                                  1                   2             3
 - Bei "Adress-Ende":                   Beginne Am Anfang dieser Adresse | Beginne bei 0 / 7 | Shift Links | Weiter in nächste Adresse |  
 - Bei Ende letzte Adrese (Zeilenende): Beginne Am Anfang dieser Adresse | Beginne bei 0 / 7 | Shift Links |
 - Schreiben über das Ende eines Devices (über das Ende der Device-Kette)ermöglichen (Position 254??)

 */
#pragma once

#if defined(__AVR__)
#include <avr/pgmspace.h>
#elif defined(ESP8266)
#include <pgmspace.h>
#endif

#if (ARDUINO >= 100)
#include <Arduino.h>
//#include <WProgram.h>
#else
#endif

#include "Print.h"

  /*!
    \enum optionEndOfDevice
   
    This enumerated type is used to defined the options
    when the cursor reaches the end of a device.
    Currently this enumeration has global scope as it can be used externally 
    with the public member function setEndOfDevice
  */
enum optionEndOfDevice {
  THIS_DEVICE,     /*!< jump to digit 7 (= position 0 on this device)*/
  FIRST_DEVICE,    /*!< jump to address 0, digit 7 (= position 0) */
  NEXT_DEVICE,     /*!< jump to address n+1 digit 7 */
  SHIFT_LEFT       /*!< scroll display - rfu - not implemented */
  };

/*
 * Segments to be switched on for characters and digits on
 * 7-Segment Displays
 */
const static uint8_t charTable [] PROGMEM  = {
  0b01111110, 0b00110000, 0b01101101, 0b01111001, 0b00110011, 0b01011011, 0b01011111, 0b01110000, //0 1 2 3 4 5 6 7   @000
  0b01111111, 0b01111011, 0b01100011, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, //8 9 °             @008
  0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
  0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
  0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b01100011, 0b00000000, 0b01000000, //          °   '   @032
  0b01001110, 0b01111000, 0b01100011, 0b00000000, 0b10000000, 0b00000001, 0b10000000, 0b01100101, //( ) °   . - . ?   @040
  0b01111110, 0b00110000, 0b01101101, 0b01111001, 0b00110011, 0b01011011, 0b01011111, 0b01110000, //0 1 2 3 4 5 6 7   @048
  0b01111111, 0b01111011, 0b01100011, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, //8 9 °             @056
  0b00000000, 0b01110111, 0b01111111, 0b01001110, 0b01111110, 0b01001111, 0b01000111, 0b01011110, //  A B C D E F G   @064
  0b00110111, 0b00110000, 0b00111100, 0b00000000, 0b00001110, 0b01110110, 0b01110110, 0b01111110, //H I J   L N N O   @072
  0b01100111, 0b11111110, 0b00000101, 0b01011011, 0b01000110, 0b00111110, 0b00111110, 0b00000000, //P Q r S T U U     @080
  0b00000000, 0b00100111, 0b01101101, 0b01001110, 0b00000000, 0b01111000, 0b00000000, 0b00001000, //  Y Z [   ]   _   @088
  0b00000000, 0b10011101, 0b00011111, 0b00001101, 0b00111101, 0b01001111, 0b01000111, 0b01011110, //  a b c d e f G   @096
  0b00010111, 0b00010000, 0b00111100, 0b00000000, 0b00001100, 0b00010101, 0b00010101, 0b00011101, //h i J   l n n o   @104
  0b01100111, 0b11111110, 0b00000101, 0b01011011, 0b01110000, 0b00011100, 0b00011100, 0b00000000, //P Q r S t u u     @112
  0b00000000, 0b00100111, 0b01101101, 0b01001110, 0b01100000, 0b01111000, 0b00000000, 0b00000000  //  Y z { | }       @120
};

/** \brief Base class for MAX7219/MAX7221

    This is the class used as base for software bitbang and HW-SPI.
    It holds all common variables and member functions.
    If you create your instance (your object), don't use this base class but one of the two implementations
    \see LedControl
    \see LedControlSpi
*/
class LedControl_base : public Print {
  protected:
      //the opcodes for the MAX7221 and MAX7219
      static constexpr byte OP_NOOP        = 0;
      static constexpr byte OP_DIGIT0      = 1;
      static constexpr byte OP_DIGIT1      = 2;
      static constexpr byte OP_DIGIT2      = 3;
      static constexpr byte OP_DIGIT3      = 4;
      static constexpr byte OP_DIGIT4      = 5;
      static constexpr byte OP_DIGIT5      = 6;
      static constexpr byte OP_DIGIT6      = 7;
      static constexpr byte OP_DIGIT7      = 8;
      static constexpr byte OP_DECODEMODE  = 9;
      static constexpr byte OP_INTENSITY   = 10;
      static constexpr byte OP_SCANLIMIT   = 11;
      static constexpr byte OP_SHUTDOWN    = 12;
      static constexpr byte OP_DISPLAYTEST = 15;
      
      uint8_t spidata[16];                       /* The array for shifting the data to the devices */
      virtual void spiTransfer(uint8_t addr, uint8_t opcode, uint8_t data);     /* Send out a single command to the device */
      uint8_t status[64];                        /* We keep track of the led-status for all 8 devices in this array */
      uint8_t maxDevices;                        /* The maximum number of devices we use */  
      // noiasca  
      uint8_t currentPosition;                   /* current position of cursor 8 devices with 8 digits - a byte variable is enough */
      uint8_t lastPosition;                      /* last position of cursor*/
      optionEndOfDevice endOfDevice = NEXT_DEVICE;       /* option what to do if end of device is reached */   
      void handleEndOfDevice(uint8_t addr);              /* handle what should happen if position exceeds end of the device */       
      bool isInitialized = false;                // backward compatibility: if user forgets to call the begin in setup, the lib will do the call on its own
 
  public:
    /* 
       Create a new LED object
    */
    LedControl_base(uint8_t numDevices = 1);
 
    /* 
      Put this method in your setup
      initialise hardware - common parts for software shiftout and hardware SPI
      needed as virtual because it will get called by ::shutdown()  
    */       
    virtual void begin();                  
    
    /** 
       \brief get the number of devices attached of this object.
       
       \returns the number of devices (MAX7219/MAX7221) of this object
     */
    uint8_t getDeviceCount();
    
    /** 
       \brief Set the shutdown (power saving) mode for the device
     
       \param addr	The address of the display to control
       \param status	If true the device goes into power-down mode. Set to false for normal operation.
     */
    void shutdown(uint8_t addr, bool status);
    
    /**
       \brief Set the number of digits (or rows) to be displayed.
       
       See datasheet for sideeffects of the scanlimit on the brightness
       of the display.
       \param addr  address of the display to control
       \param limit number of digits to be displayed (1..8)
     */
    void setScanLimit(uint8_t addr, uint8_t limit);
    
    /** 
       \brief Set the brightness of the display.
       
       \param addr      the address of the display to control
       \param intensity the brightness of the display. (0..15)
     */
    void setIntensity(uint8_t addr, uint8_t intensity);
    
    /** 
       \brief Switch all LEDss on the display off. 
       
       \param addr  address of the display to control
     */
    void clearDisplay(uint8_t addr);
    void clearDisplay();
    
    /**
      \brief Set the status of a single LEDs.
      
      \param addr	address of the display 
      \param row	  the row of the LEDs (0..7)
      \param col	  the column of the LEDs (0..7)
      \param state	If true the led is switched on, if false it is switched off
     */
    void setLed(uint8_t addr, uint8_t row, uint8_t col, boolean state);
    
    /** 
       \brief Set all 8 LEDs in a row to a new state
     
       \param addr	address of the display
       \param row	row which is to be set (0..7)
       \param value	each bit set to 1 will light up the corresponding LED.
     */
    void setRow(uint8_t addr, uint8_t row, uint8_t value);
    
    /** 
       \brief Set all 8 LEDs in a column to a new state
       
       \param addr	address of the display
       \param col	column which is to be set (0..7)
       \param value	each bit set to 1 will light up the corresponding LED
     */
    void setColumn(uint8_t addr, uint8_t col, uint8_t value);
    
    /**
      \brief Display a hexadecimal digit on a 7-Segment Display
      
      \note use lc.print(value, HEX) instead.
     
      \param addr	address of the display
      \param digit	the position of the digit on the display (0..7)
      \param value	the value to be displayed. (0x00..0x0F) 
      \param dp	sets the decimal point.
     */
    void setDigit(uint8_t addr, uint8_t digit, uint8_t value, boolean dp);
    
    /** 
      \brief Display a character on a 7-Segment display.
      
       There are only a few characters that make sense here :
      	'0','1','2','3','4','5','6','7','8','9','0',
        'A','b','c','d','E','F','H','L','P',
        '.','-','_',' ' 
        
       \note use print instead.
      
       \param addr	address of the display
       \param digit	the position of the character on the display (0..7)
       \param value	the character to be displayed. 
       \param dp	sets the decimal point.
     */
    void setChar(uint8_t addr, uint8_t digit, char value, boolean dp);
    
    // *** noiasca's new methods ***
     /** 
       \brief Change the option of what should happen at the End of the Line 
             (End of one Device/Adress)
      \param newOption one of following values 
      \see optionEndOfDevice
     */       
    void setEndOfDevice(optionEndOfDevice newOption);
    
     /** 
       \brief Set the cursor to a specific position 
       
       \param newPosition     new absolute position
     */        
    void setCursor(uint8_t newPosition);
    
     /** 
       \brief Set the Cursor to a specific position 
       
       \param addr          adress of device
       \param devPosition   position in device. The order is 01234567 !
     */ 
    void setCursor(uint8_t addr, uint8_t devPosition);
    
    size_t write(uint8_t); // noiasca
    using Print::write;    // noiasca
};
